/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <argp.h>
#include "DIS.h"
#include "IMG.h"
#include "config.h"

#ifndef MAX_INT
#define MAX_INT 32767
#endif

typedef struct GRAPH_NODE_
{
	int start;
	int end;
	double distance;
}
GRAPH_NODE;

typedef struct GRAPH_NODE_ARRAY_
{
	int node_total;
	int array_size;
	GRAPH_NODE* node;
}
GRAPH_NODE_ARRAY;

GRAPH_NODE_ARRAY
CLUS_graph_node_array_create(int array_size)
{
	GRAPH_NODE_ARRAY graph;
	
	graph.node_total=0;
	graph.array_size=array_size;
	graph.node=(GRAPH_NODE*) malloc(sizeof(GRAPH_NODE)*
		graph.array_size);
	if(graph.node==NULL)
	{
		fprintf(stderr,"Not enough memory to create graph\n");
		exit(-1);
	}
	return(graph);
}

int
CLUS_graph_node_array_destroy(GRAPH_NODE_ARRAY* graph)
{
	graph->node_total=0;
	graph->array_size=0;
	free(graph->node);
	graph->node=NULL;
	return(0);
}


GRAPH_NODE_ARRAY
CLUS_graph_node_array_copy(GRAPH_NODE_ARRAY* graph_source)
{
	GRAPH_NODE_ARRAY graph_copy;
	int i;

	graph_copy.node_total=graph_source->node_total;
	graph_copy.array_size=graph_source->node_total;
	graph_copy.node=(GRAPH_NODE*) malloc(sizeof(GRAPH_NODE)*
		graph_copy.array_size);
	if(graph_copy.node==NULL)
	{
		fprintf(stderr,"Not enough memory for a graph copy\n");
		exit(-1);
	}
	for(i=0;i<graph_copy.node_total;i++)
		graph_copy.node[i]=graph_source->node[i];

	return(graph_copy);
}

int
CLUS_graph_node_array_add(GRAPH_NODE_ARRAY* graph,
				int node_start,int node_end,
				double distance)
{
	if(graph->node_total>=graph->array_size)
	{
		graph->array_size*=2;
		graph->node=(GRAPH_NODE*) realloc(graph->node,
			graph->array_size*sizeof(GRAPH_NODE));
		if(graph->node==NULL)
		{
			fprintf(stderr,"Not enough memory to add node\n");
			exit(-1);
		}
	}
	graph->node[graph->node_total].start=node_start;
	graph->node[graph->node_total].end=node_end;
	graph->node[graph->node_total].distance=distance;
	graph->node_total++;
	return(1);
}

int CLUS_graph_node_distance_compare(void const *a,void const *b)
{
  if(((GRAPH_NODE*)a)->distance<((GRAPH_NODE*)b)->distance)
    return(-1);
  if(((GRAPH_NODE*)a)->distance>((GRAPH_NODE*)b)->distance)
    return(1);

  return(0);
}

int
CLUS_graph_node_array_sort_down(GRAPH_NODE_ARRAY* graph)
{
  qsort(graph->node,graph->node_total,sizeof(GRAPH_NODE),
	CLUS_graph_node_distance_compare);
  return(1);
}

GRAPH_NODE_ARRAY
CLUS_graph_node_array_from_disdat(DISDAT disdat)
{
  GRAPH_NODE_ARRAY graph;
  int disrow_nr;

  graph=CLUS_graph_node_array_create(disdat.counter);

  for(disrow_nr=0;disrow_nr<disdat.counter;disrow_nr++)
    CLUS_graph_node_array_add(&graph,
			      disdat.row[disrow_nr].test_nr,
			      disdat.row[disrow_nr].train_nr,
			      disdat.row[disrow_nr].disfield[0]);

  return(graph);
}


int
CLUS_graph_node_array_print(GRAPH_NODE_ARRAY* graph)
{
  int node_nr;
  fprintf(stdout,"#graph_node_array with %d nodes\n",graph->node_total);
  for(node_nr=0;node_nr<graph->node_total;node_nr++)
    fprintf(stdout,"#%d-->%d distance:%f\n",graph->node[node_nr].start,
	    graph->node[node_nr].end,graph->node[node_nr].distance);

  return(1);
}

double
CLUS_graph_node_array_get_node_distance(GRAPH_NODE_ARRAY* graph,
					int node_start,int node_end)
{
  int node_nr;

  for(node_nr=0;node_nr<graph->node_total;node_nr++)
    if(graph->node[node_nr].start==node_start
       && graph->node[node_nr].end==node_end)
      return (graph->node[node_nr].distance);

  return(-1.0);
}

GRAPH_NODE_ARRAY
CLUS_graph_node_array_get_MST_cluster(GRAPH_NODE_ARRAY* graph_source,
				      int max_cluster_total,double max_distance)
{
	GRAPH_NODE_ARRAY graph_MST;
	GRAPH_NODE_ARRAY graph_copy;
	int point_nr;
	int* point_array;
	int* point_missing_array;
	int node_nr;
	int node_start;
	int node_end;
	int max_point_nr;
	int min_point_nr;
	int MST_get_all_points;
	int cluster_total;

	graph_copy=CLUS_graph_node_array_copy(graph_source);

	CLUS_graph_node_array_sort_down(&graph_copy);

	max_point_nr=-1;
	min_point_nr=MAX_INT;
	for(node_nr=0;node_nr<graph_copy.node_total;node_nr++)
	{
	  node_start=graph_copy.node[node_nr].start;
	  if(max_point_nr<node_start)
	    max_point_nr=node_start;
	  if(min_point_nr>node_start)
	    min_point_nr=node_start;
	  node_end=graph_copy.node[node_nr].end;
	  if(max_point_nr<node_end)
	    max_point_nr=node_end;
	  if(min_point_nr>node_end)
	    min_point_nr=node_end;
	}
	if(max_point_nr<0 || min_point_nr<0)
	{
		fprintf(stderr,"CLUS_graph_node_array_get_MST can't create MST\n");
		exit(-1);
	}
	point_array=(int *) malloc(sizeof(int)*(max_point_nr+1));
	point_missing_array=(int *) malloc(sizeof(int)*(max_point_nr+1));
	if(point_array==NULL || point_missing_array==NULL)
	{
		fprintf(stderr,"Not enough memory for point_list\n");
		exit(-1);
	}
	for(point_nr=0;point_nr<=max_point_nr;point_nr++)
	  {
	    point_array[point_nr]=-1;	    
	    point_missing_array[point_nr]=0;
	  }
	for(node_nr=0;node_nr<graph_copy.node_total;node_nr++)
	  {
	    node_start=graph_copy.node[node_nr].start;
	    point_array[node_start]=node_start;
	    point_missing_array[node_start]=1;
	    node_end=graph_copy.node[node_nr].end;
	    point_array[node_end]=node_end;
	    point_missing_array[node_end]=1;
	  }
	
	graph_MST=CLUS_graph_node_array_create(graph_copy.node_total);

	for(node_nr=0;node_nr<graph_copy.node_total;node_nr++)
	{
		node_start=graph_copy.node[node_nr].start;
		node_end=graph_copy.node[node_nr].end;
		if(graph_copy.node[node_nr].distance>=max_distance)
		  {
		    for(point_nr=min_point_nr;point_nr<=max_point_nr;point_nr++)
		      if(point_missing_array[point_nr]==1)
			CLUS_graph_node_array_add(&graph_MST,point_nr,point_nr,0.0);	    
		    node_nr=graph_copy.node_total;
		  }
		else
		  {
		    if(point_array[node_start]!=point_array[node_end])
		      {
			CLUS_graph_node_array_add(&graph_MST,
						  node_start,node_end,
						  graph_copy.node[node_nr].distance);
			point_missing_array[node_start]=0;
			point_missing_array[node_end]=0;
			if(point_array[node_start]<point_array[node_end])
			  {
			    node_start=point_array[node_start];
			    node_end=point_array[node_end];
			  }
			else
			  {    
			    point_nr=point_array[node_end];
			    node_end=point_array[node_start];
			    node_start=point_nr;
			  }
			
			MST_get_all_points=1;
			cluster_total=0;
			for(point_nr=min_point_nr;point_nr<=max_point_nr;point_nr++)
			  {
			    if(point_array[point_nr]==node_end)
			      point_array[point_nr]=node_start;
			    if(point_array[point_nr]>min_point_nr)
			      MST_get_all_points=0;
			    if(point_array[point_nr]==point_nr)
			      cluster_total++;
			  }
			if(MST_get_all_points || cluster_total<=max_cluster_total)
			  {
			    for(point_nr=min_point_nr;point_nr<=max_point_nr;point_nr++)
			      if(point_missing_array[point_nr]==1)
				CLUS_graph_node_array_add(&graph_MST,point_nr,point_nr,0.0);		
			    node_nr=graph_copy.node_total;
			  }
		      }
		  }
	}
	CLUS_graph_node_array_destroy(&graph_copy);
	free(point_array);
	return(graph_MST);
}

int**
CLUS_graph_node_array_to_cluster_array(GRAPH_NODE_ARRAY* graph)
{
  int** cluster_array;  
  int point_nr;
  int* point_array;  
  int* point_array_temp;
  int max_point_nr;
  int min_point_nr;
  int node_nr;
  int node_start;
  int node_end;
  int cluster_total;
  int cluster_min_point;
  int cluster_element_nr;
  int point_nr_temp;
  int cluster_nr;

  max_point_nr=-1;
  min_point_nr=MAX_INT;
  for(node_nr=0;node_nr<graph->node_total;node_nr++)
    {
      node_start=graph->node[node_nr].start;
      if(max_point_nr<node_start)
	max_point_nr=node_start;
      if(min_point_nr>node_start)
	min_point_nr=node_start;
      node_end=graph->node[node_nr].end;
      if(max_point_nr<node_end)
	max_point_nr=node_end;
      if(min_point_nr>node_end)
	min_point_nr=node_end;
    }
  if(max_point_nr<0 || min_point_nr<0)
    {
      fprintf(stderr,"CLUS_graph_node_array_to_cluster_array can't create point_list\n");
      exit(-1);
    }
  point_array=(int *) malloc((max_point_nr+1)*sizeof(int));
  point_array_temp=(int *) malloc((max_point_nr+1)*sizeof(int));
  if(point_array==NULL || point_array_temp==NULL)
    {
      fprintf(stderr,"Not enough memory for point_array\n");
      exit(-1);
    }
  for(point_nr=0;point_nr<=max_point_nr;point_nr++)
    {
      point_array[point_nr]=-1;
      point_array_temp[point_nr]=0;      
    }
  for(node_nr=0;node_nr<graph->node_total;node_nr++)
    {
      node_start=graph->node[node_nr].start;
      if( point_array[node_start]!=-1 )
	node_start=point_array[node_start];
      node_end=graph->node[node_nr].end;
      if( point_array[node_end]!=-1 )
	node_end=point_array[node_end];

      if(node_start>node_end)
	{
	  point_nr=node_start;
	  node_start=node_end;
	  node_end=point_nr;
	}
      point_array[node_end]=node_start;
      point_array[node_start]=node_start;

      for(point_nr=min_point_nr;point_nr<=max_point_nr;point_nr++)
	if(point_array[point_nr]==node_end)
	  point_array[point_nr]=node_start;
    }

  cluster_total=0;
  for(point_nr=min_point_nr;point_nr<=max_point_nr;point_nr++)
    {
      cluster_min_point=point_array[point_nr];
      if(cluster_min_point!=-1)
	{
	  if(point_array_temp[cluster_min_point]==0)
	    cluster_total++;
	  point_array_temp[cluster_min_point]++;
	}
    }

  //  printf("cluse: %d\n",cluster_total);

  cluster_array=(int**) malloc((cluster_total+1)*sizeof(int*));
  if(cluster_array==NULL)
    {
      fprintf(stderr,"Not enough memory for cluster_array\n");
      exit(-1);
    }
  cluster_array[cluster_total]=NULL;

  cluster_nr=0;
  for(point_nr_temp=min_point_nr;point_nr_temp<=max_point_nr;point_nr_temp++)
    {  
      if(point_array_temp[point_nr_temp]>0)
	{
	  cluster_array[cluster_nr]=(int*) malloc((point_array_temp[point_nr_temp]+1)*sizeof(int));
	  if(cluster_array[cluster_nr]==NULL)
	    {
	      fprintf(stderr,"Not enough memory for cluster_array\n");
	      exit(-1);
	    }
	  cluster_element_nr=0;
	  for(point_nr=min_point_nr;point_nr<=max_point_nr;point_nr++)
	    {
	      if(point_nr_temp==point_array[point_nr])
		{
		  cluster_array[cluster_nr][cluster_element_nr]=point_nr;
		  cluster_element_nr++;
		}
	      cluster_array[cluster_nr][cluster_element_nr]=-2;
	    }
	  cluster_nr++;
	}
    }
  free(point_array);
  free(point_array_temp);
  
  return(cluster_array);
}

int
CLUS_cluster_array_print(int** cluster_array)
{
  int cluster_nr;
  int cluster_element_nr;

  for(cluster_nr=0;cluster_array[cluster_nr]!=NULL;cluster_nr++);
  fprintf(stdout,"#cluster_array with %d clusters\n",cluster_nr);
  for(cluster_nr=0;cluster_array[cluster_nr]!=NULL;cluster_nr++)
    {
      fprintf(stdout,"#cluster %d:",cluster_nr);
      for(cluster_element_nr=0;cluster_array[cluster_nr][cluster_element_nr]>=0;cluster_element_nr++)
	fprintf(stdout," %d",cluster_array[cluster_nr][cluster_element_nr]);
      fprintf(stdout,"\n");
    }

  return(1);
}

int*
CLUS_cluster_array_get_center(int** cluster_array,GRAPH_NODE_ARRAY* graph)
{
  int* image_nr_array;
  int cluster_nr;
  int cluster_element_nr;
  int cluster_element_nr_b;
  int cluster_total;
  double distance_sum;
  double best_distance_sum;
  int best_image_nr;
  double distance;

  for(cluster_total=0;cluster_array[cluster_total]!=NULL;cluster_total++);
  image_nr_array= (int*) malloc((cluster_total+1)*sizeof(int));
  if(image_nr_array==NULL)
    {
      fprintf(stderr,"Not enough memory for image_nr_array\n");
      exit(-1);
    }  
  image_nr_array[cluster_total]=-2;

  for(cluster_nr=0;cluster_array[cluster_nr]!=NULL;cluster_nr++)
    {
      best_distance_sum=DBL_MAX;
      best_image_nr=-1;
      fprintf(stdout,"#cluster %d:",cluster_nr);
      for(cluster_element_nr=0;cluster_array[cluster_nr][cluster_element_nr]>=0;cluster_element_nr++)
	{
	  distance_sum=0.0;
	  fprintf(stdout," %d",cluster_array[cluster_nr][cluster_element_nr]);
	  for(cluster_element_nr_b=0;
	      cluster_array[cluster_nr][cluster_element_nr_b]>=0;cluster_element_nr_b++)
	    {
	      if(cluster_array[cluster_nr][cluster_element_nr]!=
		 cluster_array[cluster_nr][cluster_element_nr_b])
		{
		  distance=CLUS_graph_node_array_get_node_distance(graph,
				     cluster_array[cluster_nr][cluster_element_nr_b],
				     cluster_array[cluster_nr][cluster_element_nr]);
		  //test_images_distances-->train_image_distance
		  if(distance<0.0)
		    {
		      fprintf(stderr,"Can't find node %d-->%d\n",
			      cluster_array[cluster_nr][cluster_element_nr_b],
			      cluster_array[cluster_nr][cluster_element_nr]);
		      //		      exit(-1);
		      distance=0;
		    }
		  distance_sum+=distance;
		}
	    }
	  fprintf(stdout,",%f",distance_sum);
	  if(distance_sum<best_distance_sum)
	    {
	      best_distance_sum=distance_sum;
	      best_image_nr=cluster_array[cluster_nr][cluster_element_nr];
	    }
	}
      fprintf(stdout,"\n");
      image_nr_array[cluster_nr]=best_image_nr;
    }


  return(image_nr_array);
}

const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char *disdat_name;            /* ARG1 */
  int test_class;
  int train_class;
  int max_cluster_total;
  double max_distance;
  int verbose;
  int disfield_nr;
  int sym_dis;
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"verbose", 'v', 0, 0, "Produce verbose output"},
  {"maxcluster",'C',"N", 0, "Maximal clusters to split imagenumbers"},
  {"dislimit", 'D', "N", 0, "Distance limit for the clusters"},
  //  {"trainclass",'H', "N", 0, "Class for train images"},
  //  {"testclass",'G',"N",0, "Class for test images"},  
  {"disfield",'f',"N",0,"Select the N'th disfield, counted from 0"},
  {"symdis",'y',0,0, "Symetric distance"},
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'v':
      arguments->verbose = 1;
      break;
    case 'C':
      if(sscanf(arg,"%d%s",&(arguments->max_cluster_total),arg)!=1 
	 || arguments->max_cluster_total<1)
	argp_usage(state); 
      break;
    case 'D':
      if(sscanf(arg,"%lf%s",&(arguments->max_distance),arg)!=1
	 || arguments->max_distance<0)
	argp_usage(state); 
      break;
    case 'G':
      if (sscanf (arg, "%d%s", &(arguments->test_class), arg) != 1)
	  argp_usage(state);
      break;
    case 'H':
      if (sscanf (arg, "%d%s", &(arguments->train_class), arg) != 1)
	  argp_usage(state);
      break;
    case 'f':
      if(sscanf(arg,"%d%s",&(arguments->disfield_nr),arg)!=1
	 || arguments->disfield_nr<0)
	argp_usage(state); 
      break;
    case 'y':
      arguments->sym_dis=1;
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 1)
	{
	  argp_usage(state);
	}
      arguments->disdat_name = arg;
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 1)
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "FILE";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"cluster -- The program read a disdat file and create clusters from the imagenumbers with the minimal spanning tree.\v When FILE is -, read standard input.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

int main(int argc,char* argv[])
{
  struct arguments arguments;
  GRAPH_NODE_ARRAY graph,graph_MST;
  DISDAT disdat;
  DISDAT disdat_copy;
  int** cluster_array;
  int* image_nr_array;
  
 /* Set argument defaults */
  arguments.verbose = 0;
  arguments.disdat_name = "";
  arguments.disfield_nr = 0;
  arguments.test_class=-1;
  arguments.train_class=-1;
  arguments.max_cluster_total=1;
  arguments.max_distance=DBL_MAX;
  arguments.sym_dis=0;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  disdat=DIS_read_disdat_file(arguments.disdat_name);


  if(arguments.disfield_nr!=0)
    {
      DIS_disdat_swap_disfield(&disdat,arguments.disfield_nr);
    }

  if(arguments.sym_dis)
    {
      disdat_copy=DIS_disdat_copy(disdat);
      DIS_disdat_sym_distance(&disdat);
    }

  graph=CLUS_graph_node_array_from_disdat(disdat);

  if(arguments.verbose)
    CLUS_graph_node_array_print(&graph);
 
  graph_MST=CLUS_graph_node_array_get_MST_cluster(&graph,
						  arguments.max_cluster_total,
						  arguments.max_distance);

  if(arguments.verbose)
    CLUS_graph_node_array_print(&graph_MST);

  cluster_array=CLUS_graph_node_array_to_cluster_array(&graph_MST);
 
  // if(arguments.verbose)
  //  CLUS_cluster_array_print(cluster_array);

  if(arguments.sym_dis)
    { 
      CLUS_graph_node_array_destroy(&graph);
      graph=CLUS_graph_node_array_from_disdat(disdat_copy);
      DIS_disdat_free(disdat_copy);
      image_nr_array=CLUS_cluster_array_get_center(cluster_array,&graph);
    }
  else
    image_nr_array=CLUS_cluster_array_get_center(cluster_array,&graph);
  
  IMG_image_nr_array_print(image_nr_array);

  CLUS_graph_node_array_destroy(&graph);
  CLUS_graph_node_array_destroy(&graph_MST);
  DIS_disdat_free(disdat);
  return(0);
}







