/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <argp.h>
#include <float.h>
#include <math.h>
#include <gd.h>
#include <gdfontl.h>
#include "IMG.h"
#include "config.h"

#define IMG_red_value 1
#define IMG_green_value 2
#define IMG_blue_value 4
#define IMG_grey_value 7


const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char *args[2];            /* ARG1 */
  int img_depth_list[5];
  int img_depth;
  int verbose;
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"red",'R', 0, 0, "Use red image values"},
  {"green",'G', 0, 0, "Use green image values"},
  {"blue",'B', 0, 0, "Use blue image values"},
  {"grey",'E', 0, 0, "Use grey image values"},
  {"verbose", 'v', 0, 0, "Verbose output"},
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'R':
      arguments->img_depth_list[arguments->img_depth] = IMG_red_value;
      arguments->img_depth++;
      break;
    case 'G':
      arguments->img_depth_list[arguments->img_depth] = IMG_green_value;
      arguments->img_depth++;
      break;
    case 'B':
      arguments->img_depth_list[arguments->img_depth] = IMG_blue_value;
      arguments->img_depth++;
      break;
    case 'E':
      arguments->img_depth_list[arguments->img_depth] = IMG_grey_value;
      arguments->img_depth++;
      break;
    case 'v':
      arguments->verbose=1;
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 2)
	{
	  argp_usage(state);
	}
      arguments->args[state->arg_num] = arg;
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 2)
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1] [FILE2]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"images2jf -- The program read FILE1 which contain the imagefilenames, and write FILE2.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};
	
int
IMG_image_file2img (char *filename, int* img_depth_list, IMG* img)
{
  gdImagePtr gdimg;
  FILE *image_file;
  int x, y;
  int colorindex;
  double *tempdata;
  int img_depth;
  int img_file_type;

  img_file_type=-1;
  if(strcmp(".png",&(filename[strlen(filename)-4]))==0)
    img_file_type=1;
  else
    if(strcmp(".jpg",&(filename[strlen(filename)-4]))==0) 
      img_file_type=2;
    else
      {
	fprintf(stderr,"filename %s has to end with .jpg or .png\n",filename);
	exit(-1);
      }

  if ((image_file = fopen (filename, "rb")) == NULL)
    {
      fprintf(stderr,"can't open file %s\n",filename);
      exit (0);
    }

  if(img_file_type==1)
    gdimg = gdImageCreateFromPng (image_file);
  if(img_file_type==2)
    gdimg = gdImageCreateFromJpeg (image_file);
  
  fclose (image_file);

  img->x_size=gdImageSX (gdimg);
  img->y_size=gdImageSY (gdimg);

  for(img_depth=0;img_depth_list[img_depth]>=0;img_depth++)
    {
      tempdata=(double*) malloc(img->x_size * img->y_size * sizeof(double));
      if(tempdata==NULL)
	{
	  fprintf(stderr,"not enough memory for IMG_image_file2img in file %s\n",filename);
	  exit (0);
	}      
      
      for (y = 0; y < img->y_size; y++)
	for (x = 0; x < img->x_size; x++)
	  {
	    colorindex = gdImageGetPixel (gdimg, x, y);
	    switch(img_depth_list[img_depth])
	      {
	      case IMG_grey_value:
		tempdata[(img->x_size * y) + x] = (gdimg->red[colorindex] +
						   gdimg->green[colorindex] +
						   gdimg->blue[colorindex]) / 3;
		break;
	      case IMG_red_value:
		tempdata[(img->x_size * y) + x] = gdimg->red[colorindex];
		break;
	      case IMG_green_value:
		tempdata[(img->x_size * y) + x] = gdimg->green[colorindex];
		break;
	      case IMG_blue_value:
		tempdata[(img->x_size * y) + x] = gdimg->blue[colorindex];
		break;
	      }
	  }
      img->data_list[img_depth]=tempdata;
      img->data_length[img_depth]=img->x_size*img->y_size;
      img->dis_mode[img_depth]=0;
    }
  img->class=-1;
  sscanf (filename, "%d", &(img->class));
  
  gdImageDestroy(gdimg);
    
  return (1);
}

#define BUFFER_LENGTH 200
IMG_list* IMG_images_text_list_to_img_list (char *images_list_name,int* img_depth_list)
{
  IMG_list* img_list;
  int img_nr;
  FILE *images_list_file;
  char buffer[BUFFER_LENGTH];
  int line;
  int img_depth;


  for(img_depth=0;img_depth_list[img_depth]>=0;img_depth++);

  if ((images_list_file = fopen (images_list_name, "r")) == NULL)
    {
      fprintf (stderr, "can't open file %s\n", images_list_name);
      exit (0);
    }
  
  line=0;
  while (1)
    {
      fgets (buffer, BUFFER_LENGTH, images_list_file);
      if (feof (images_list_file) != 0)
	break;
      line++;
      if (strlen (buffer) + 2 > BUFFER_LENGTH)
	{
	  printf ("line %d: filename to long (max. %d)\n",
		  line, BUFFER_LENGTH - 2);
	  exit (0);
	}
    }
  img_nr=line;
  
  img_list = IMG_create_new_list(img_nr, img_depth+1);
  rewind(images_list_file);
  line=0;
  while (1)
    {
      fgets (buffer, BUFFER_LENGTH, images_list_file);
      if (feof (images_list_file) != 0)
	break;
      line++;
      if(line>img_nr)
	{
	  printf ("the file %s must be modified during reading\n",
		  images_list_name);
	  exit (0);
	}	  
      if (strlen (buffer) + 2 > BUFFER_LENGTH)
	{
	  printf ("line %d: filename to long (max. %d)\n",
		  line, BUFFER_LENGTH - 2);
	  exit (0);
	}
      buffer[strlen (buffer) - 1] = 0;
      IMG_image_file2img (buffer,img_depth_list,&(img_list->img[line-1]));
    }
  fclose (images_list_file);
  return (img_list);
}


int
main (int argc, char *argv[])
{
  struct arguments arguments;
  IMG_list* img_list;
  int img_depth;
  
  /* Set argument defaults */
  arguments.args[0] = "";
  arguments.args[1] = "";
  arguments.img_depth = 0;
  arguments.verbose = 0;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  if(arguments.img_depth == 0)
    {
      arguments.img_depth_list[0]=IMG_grey_value;
      arguments.img_depth=1;
    }
  arguments.img_depth_list[arguments.img_depth]=-2;

  for(img_depth=0;arguments.img_depth_list[img_depth]>=0;img_depth++)
    switch(arguments.img_depth_list[img_depth])
      {
      case IMG_grey_value:
	fprintf(stdout,"#img_depth %d: grey values\n",img_depth);
	break;
      case IMG_red_value:
	fprintf(stdout,"#img_depth %d: red values\n",img_depth);	
	break;
      case IMG_green_value:
	fprintf(stdout,"#img_depth %d: green values\n",img_depth);
	break;
      case IMG_blue_value:
	fprintf(stdout,"#img_depth %d: blue values\n",img_depth);
	break;
      }

  img_list = IMG_images_text_list_to_img_list (arguments.args[0],arguments.img_depth_list);

  IMG_save_list(img_list,arguments.args[1]);
  if(arguments.verbose)
    IMG_print_list_info (img_list); 
  IMG_free_list(img_list);

  exit (1);
}
