/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <argp.h>
#include <math.h>
#include "config.h"
#ifdef USE_SPLINElib
#include <coeff.h>
#include <interpol.h>
#endif
#include "IMG.h"

const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char *args[2];            /* ARG1 */
  int splinedegree;
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"degree", 'd', "N", 0, "Splinedegree 2-5"}, 
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'd':
      if(sscanf(arg,"%d%s",&(arguments->splinedegree),arg)!=1
	 || arguments->splinedegree<2 || arguments->splinedegree>5)
	argp_usage(state); 
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 2)
	{
	  argp_usage(state);
	}
      arguments->args[state->arg_num] = arg;
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 2 )
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1] [FILE2]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"jfgradient -- The program reads FILE1, and writes gradient FILE2.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

int img_spline_gradient(IMG* img, IMG* grad_img, int splinedegree)
{
#ifdef USE_SPLINElib
  float* spline_img;
  int i;
  int x,y;
  int img_depth;
  double* tempdata;
  double** data_list;
  int** data_length;

  spline_img=(float*) malloc(sizeof(float)*img->x_size*img->y_size);
  if(spline_img==NULL)
    {
      printf("Not enough memory for spline_img\n");
      exit(-1);
    }

  for(img_depth=0;img_depth<=IMG_get_depth(img);img_depth++)
    {
      if(img->data_length[0]==
	 img->x_size*img->y_size)
	{
	  for(i=0;i<img->x_size*img->y_size;i++)
	    spline_img[i]=(float) img->data_list[0][i];
	  
	  if (SamplesToCoefficients(spline_img, img->x_size, img->y_size, splinedegree)) {
	    printf("Change of basis failed\n");
	    exit(-1); 
	  }

	  tempdata=(double*) malloc(sizeof(double)*img->x_size*img->y_size);
	  if(tempdata==NULL)
	    {
	      printf("Not enough memory for spline_img\n");
	      exit(-1);
	    }  

	  for(y=0;y<img->y_size;y++)
	    for(x=0;x<img->x_size;x++)
	      tempdata[x+(y*img->x_size)]=
		(double) spline_img[((x+1)%img->x_size)+(y*img->x_size)] -
		(double) spline_img[((img->x_size+x-1)%img->x_size)+(y*img->x_size)];	      

	  grad_img->data_list[img_depth*2]=tempdata;
	  grad_img->data_length[img_depth*2]=img->x_size*img->y_size;
	  grad_img->dis_mode[img_depth*2]=0;

	  tempdata=(double*) malloc(sizeof(double)*img->x_size*img->y_size);
	  if(tempdata==NULL)
	    {
	      printf("Not enough memory for spline_img\n");
	      exit(-1);
	    }  

	  for(y=0;y<img->y_size;y++)
	    for(x=0;x<img->x_size;x++)
	      tempdata[x+(y*img->x_size)]=
		(double) spline_img[x+(((y+1)%img->y_size)*img->x_size)] -
		(double) spline_img[x+(((img->y_size+y-1)%img->y_size)*img->x_size)];

	  grad_img->data_list[(img_depth*2)+1]=tempdata;
	  grad_img->data_length[(img_depth*2)+1]=img->x_size*img->y_size;
	  grad_img->dis_mode[(img_depth*2)+1]=0;
	}
      else
	{
	  grad_img->data_list[img_depth*2]=NULL;
	  grad_img->data_length[img_depth*2]=0;
	  grad_img->data_list[(img_depth*2)+1]=NULL;
	  grad_img->data_length[(img_depth*2)+1]=0;
	}
    }
  grad_img->dis_mode[img_depth*2]=-2;
  free(spline_img);
#endif
#ifndef USE_SPLINElib
  fprintf(stderr,"Not compiled with spline support.\n");
#endif
  return(0);
}


int main (int argc, char **argv)
{
  struct arguments arguments;
  IMG_list* img_list;
  IMG_list* img_list_new;
  int img_nr;
  
  /* Set argument defaults */
  arguments.args[0] = "";
  arguments.args[1] = "";
  arguments.splinedegree=3;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  img_list=IMG_load_list (arguments.args[0]);

  img_list_new =
    IMG_create_empty_list (img_list, (IMG_get_depth_list(img_list)+1)*2);
  fprintf(stdout,"#use a splinedegree of %d\n",arguments.splinedegree);

  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    img_spline_gradient( &(img_list->img[img_nr]), &(img_list_new->img[img_nr]), arguments.splinedegree);

  IMG_save_list(img_list_new,arguments.args[1]);
  IMG_free_list(img_list);
  IMG_free_list(img_list_new);
  return 0;
}










