/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <argp.h>
#include "IMG.h"
#include "config.h"

#define norm_WITHOUT_DEPTH 0
#define norm_WITH_DEPTH 1
#define norm_EACH_IMAGE 2


const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char *args[2];            /* ARG1 */
  double newstart;
  double newend;
  double gamma;
  int mode;
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"start", 's', "R", 0, "New startgreyvalue"}, 
  {"end",'e', "R", 0, "New endgreyvalue"},
  {"gamma",'g',"R",0, "Value for gammatransformation after greyvaluenorm"},
  {"all",'a',0,0,"Norm all images (include imagedepth)"},
  {"each",'n',0,0,"Norm each image instead whole list"},
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 's':
      if(sscanf(arg,"%lf%s",&(arguments->newstart),arg)!=1)
	argp_usage(state); 
      break;
    case 'e':
      if(sscanf(arg,"%lf%s",&(arguments->newend),arg)!=1)
	argp_usage(state); 
      break;    
    case 'g':
      if(sscanf(arg,"%lf%s",&(arguments->gamma),arg)!=1)
	argp_usage(state); 
      break;
    case 'a':
      arguments->mode+=norm_WITH_DEPTH;
      break;
    case 'n':
      arguments->mode+=norm_EACH_IMAGE;
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 2)
	{
	  argp_usage(state);
	}
      arguments->args[state->arg_num] = arg;
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 2)
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1] [FILE2]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"jfnorm -- The program reads FILE1, and norms the images and writes FILE2.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

int img_gammatrans_list(IMG_list* img_list, int field, double gamma)
{
  int img_nr;
  int i;
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    {
      for(i=0;i<img_list->img[img_nr].data_length[field];i++)
	img_list->img[img_nr].data_list[field][i]=
	  pow(img_list->img[img_nr].data_list[field][i],gamma);
    }
	
  fprintf (stdout, "$IMG_gammatrans(%s,%d,%f);\n", img_list->name,
	   field, gamma);

  return(1);
}


int
IMG_norm_each_imagedepth_form_list (IMG_list * img_list, double new_min, double new_max)
{
  int img_nr;
  int x;
  double min;
  double max;
  int img_depth;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    {
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list);img_depth++)
	{
	  max=-DBL_MAX;
	  min=DBL_MAX;
	  for (x = 0; x < img_list->img[img_nr].data_length[img_depth]; x++)
	    {
	      if (max < img_list->img[img_nr].data_list[img_depth][x])
		max = img_list->img[img_nr].data_list[img_depth][x];
	      if (min > img_list->img[img_nr].data_list[img_depth][x])
		min = img_list->img[img_nr].data_list[img_depth][x];
	    }
	  for (x = 0; x < img_list->img[img_nr].data_length[img_depth]; x++)
	    img_list->img[img_nr].data_list[img_depth][x] =
	      (((img_list->img[img_nr].data_list[img_depth][x] - min) / (max -
								     min)) *
	       (new_max - new_min)) + new_min;
	}
    }
  fprintf (stdout, "$IMG_norm_each_imagedepth_from_list(%s,%f,%f);\n", img_list->name, new_min, new_max);
  return (0);
}

int
IMG_norm_each_img_form_list (IMG_list * img_list, double new_min, double new_max)
{
  int img_nr;
  int x;
  double min;
  double max;
  int img_depth;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    {
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list);img_depth++)
	{
	  max=-DBL_MAX;
	  min=DBL_MAX;
	  for (x = 0; x < img_list->img[img_nr].data_length[img_depth]; x++)
	    {
	      if (max < img_list->img[img_nr].data_list[img_depth][x])
		max = img_list->img[img_nr].data_list[img_depth][x];
	      if (min > img_list->img[img_nr].data_list[img_depth][x])
		min = img_list->img[img_nr].data_list[img_depth][x];
	    }
	}
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list);img_depth++)
	{     
	  for (x = 0; x < img_list->img[img_nr].data_length[img_depth]; x++)
	    img_list->img[img_nr].data_list[img_depth][x] =
	      (((img_list->img[img_nr].data_list[img_depth][x] - min) / (max -
								     min)) *
	       (new_max - new_min)) + new_min;
	}
    }
  fprintf (stdout, "$IMG_norm_each_img_from_list(%s,%f,%f);\n", img_list->name, new_min, new_max);
  return (0);
}

int main (int argc, char **argv)
{
  struct arguments arguments;
  IMG_list* img_list;
  double min,max;
  int img_depth;
  
  /* Set argument defaults */
  arguments.args[0] = "";
  arguments.args[1] = "";
  arguments.newstart=0.0;
  arguments.newend=1.0;
  arguments.gamma=1.0;
  arguments.mode=norm_WITHOUT_DEPTH;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);
  
  img_list=IMG_load_list (arguments.args[0]);


  switch(arguments.mode)
    {
    case norm_WITHOUT_DEPTH:
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list);img_depth++)
	{
	  min = DBL_MAX;
	  max = -DBL_MAX;
	  IMG_get_min_max_list (img_list, img_depth, &min, &max);
	  IMG_norm_list (img_list, img_depth, min, max, arguments.newstart, arguments.newend);
	  img_gammatrans_list(img_list, img_depth, arguments.gamma);  
	}
      break;

    case norm_WITH_DEPTH:
      min = DBL_MAX;
      max = -DBL_MAX;
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list);img_depth++)
	IMG_get_min_max_list (img_list, img_depth, &min, &max);  
      
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list);img_depth++)
	{
	  IMG_norm_list (img_list, img_depth, min, max, arguments.newstart, arguments.newend);
	  img_gammatrans_list(img_list, img_depth, arguments.gamma);  
	}
      break;
   case (norm_WITHOUT_DEPTH+norm_EACH_IMAGE):
      IMG_norm_each_imagedepth_form_list(img_list,arguments.newstart,arguments.newend);
      img_gammatrans_list(img_list, img_depth, arguments.gamma);
      break;

    case (norm_WITH_DEPTH+norm_EACH_IMAGE):
      IMG_norm_each_img_form_list(img_list,arguments.newstart,arguments.newend);
      img_gammatrans_list(img_list, img_depth, arguments.gamma);  
      break;


    }

  IMG_save_list(img_list,arguments.args[1]);

 exit(1);
}
