/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <argp.h>
#include <math.h>
#include "config.h"
#ifdef USE_SPLINElib
#include <coeff.h>
#include <interpol.h>
#endif
#include "IMG.h"


#define scale_BRESENHAM 1
#define scale_SPLINE 2

const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char *args[2];            /* ARG1 */
  int mode;
  int width;
  int height;
  int splinedegree;
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"mode",'M',"bresenham|spline",0,"Choose the scale algorithm"},
  {"degree", 'd', "N", 0, "Splinedegree 2-5"}, 
  {"width", 'w', "N", 0, "Destination image width"},
  {"height", 'h', "N", 0, "Destination image height"},
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'M':
      if(!strcmp(arg,"spline"))
	arguments->mode=scale_SPLINE;
      else
	if(!strcmp(arg,"bresenham"))
	  arguments->mode=scale_BRESENHAM;	
	else
	  argp_usage(state);
      break;
    case 'w':
      if(sscanf(arg,"%d%s",&(arguments->width),arg)!=1
	 || arguments->width<1)
	argp_usage(state); 
      break;
    case 'h':
      if(sscanf(arg,"%d%s",&(arguments->height),arg)!=1
	 || arguments->height<1)
	argp_usage(state); 
      break;
    case 'd':
      if(sscanf(arg,"%d%s",&(arguments->splinedegree),arg)!=1
	 || arguments->splinedegree<2 || arguments->splinedegree>5)
	argp_usage(state); 
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 2)
	{
	  argp_usage(state);
	}
      arguments->args[state->arg_num] = arg;
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 2 || (arguments->width+arguments->height) < 1)
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1] [FILE2]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"jfscale -- The program reads FILE1, scale the images and writes FILE2.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

void scale_bresenham( double *inputdata, int srcHeight, int srcWidth, double *outputdata, int destHeight, int destWidth) {

  double x_weight, y_weight;
  int x,y,x_pos_orig, x_pos_scaled, index_orig, index_scaled,y_pos_orig, y_pos_scaled;
  double akku;
  double *work;

  work=(double*)calloc(destWidth*srcHeight,sizeof(double));
  if(work==NULL)
    {
      fprintf(stderr,"No memory for scale_bresenham");
      exit(-1);
    }

  x_weight = (double) destWidth  / (double) srcWidth;
  y_weight = (double) destHeight / (double) srcHeight;

  for (y = 0; y < srcHeight; y++) {
    x_pos_orig          = 0;
    x_pos_scaled        = 0;
    index_orig          = y * srcWidth;
    index_scaled        = y * destWidth;
    akku                        = (double) 0;
    work[index_scaled] = (double) 0;
    while (x_pos_orig < srcWidth){
      if ( (akku + x_weight) >= 1.0){
        work[index_scaled] += inputdata[index_orig] * (1.0 - akku);
        akku += x_weight - 1.0;
        x_pos_scaled++;
        index_scaled++;

        while (akku >= 1.0){
         work[index_scaled] = inputdata[index_orig];
          x_pos_scaled++;
          index_scaled++;
          akku -= 1.0;
        }
        if (x_pos_scaled < destWidth){
          work[index_scaled] = inputdata[index_orig] * (akku);
        }
      } else {
        work[index_scaled] += inputdata[index_orig] * x_weight;
        akku += x_weight;
      }
      x_pos_orig++;
      index_orig++;
    }
  }
  
  for (x = 0; x < destWidth; x++){
    y_pos_orig          = 0;
    y_pos_scaled        = 0;
    index_orig          = x;
    index_scaled        = x;
    akku                = (double) 0;
    outputdata[index_scaled] = (double) 0;

    while (y_pos_orig < srcHeight) { 
      if ( (akku + y_weight) >= 1.0) {
        outputdata[index_scaled] += (1.0 - akku) * work[index_orig];
        akku += y_weight - 1.0;
        y_pos_scaled++; 
        index_scaled += destWidth;

        while (akku >= 1.0) {
          outputdata[index_scaled] = work[index_orig];
          y_pos_scaled++; 
          index_scaled += destWidth;
          akku -= 1.0;
        }

        if (y_pos_scaled < destHeight){
          outputdata[index_scaled] = work[index_orig] * (akku);
        }
      }else {
        outputdata[index_scaled] += work[index_orig] * y_weight;
        akku += y_weight;
      }
      y_pos_orig++; 
      index_orig += destWidth;
    }
  } /* for x */
  free(work);
}

int img_scale_bresenham(IMG* img, int newHeight, int newWidth)
{
  int img_depth;
  double* tempdata;
  int i;

  for(img_depth=0;img->dis_mode[img_depth]!=-2;img_depth++)
    {
      if(img->data_length[img_depth]==
	 img->x_size*img->y_size)
	{
	  tempdata=(double*) malloc(sizeof(double)*newHeight*newWidth);
	  if(tempdata==NULL)
	    {
	      printf("Not enough memory for bresenham_img\n");
	      exit(-1);
	    }
	  scale_bresenham(img->data_list[img_depth],
			  img->y_size, img->x_size,
			  tempdata,
			  newHeight,newWidth);
	  
	  free(img->data_list[img_depth]);
	  img->data_list[img_depth]=tempdata;
	  img->data_length[img_depth]=newWidth*newHeight;
	}
    }
  img->x_size=newWidth;
  img->y_size=newHeight;
  return(0);
}

int img_scale_spline(IMG* img, int newHeight, int newWidth, int splinedegree)
{
#ifdef USE_SPLINElib
  float* spline_img;
  int i;
  int x,y;
  int img_depth;
  double* tempdata;

  spline_img=(float*) malloc(sizeof(float)*img->x_size*img->y_size);
  if(spline_img==NULL)
    {
      printf("Not enough memory for spline_img\n");
      exit(-1);
    }
  
  for(img_depth=0;img->dis_mode[img_depth]!=-2;img_depth++)
    {
      if(img->data_length[img_depth]==
	 img->x_size*img->y_size)
	{
	  tempdata=(double*) malloc(sizeof(double)*newWidth*newHeight);
	  if(tempdata==NULL)
	    {
	      printf("Not enough memory for spline_img\n");
	      exit(-1);
	    }

	  for(i=0;i<img->x_size*img->y_size;i++)
	    spline_img[i]=(float) img->data_list[img_depth][i];
	  
	  if (SamplesToCoefficients(spline_img, img->x_size, img->y_size, splinedegree)) {
	    printf("Change of basis failed\n");
	    exit(-1);
	  } 
	  
	  for(y=0;y<newHeight;y++)
	    for(x=0;x<newWidth;x++)
	      tempdata[(y*newWidth)+x]=InterpolatedValue(spline_img,img->x_size, img->y_size,
							 ((x)*((double)img->x_size-1.0)/((double)newWidth-1.0)),
							 ((y)*((double)img->y_size-1.0)/((double)newHeight-1.0)),
							 splinedegree);
	  free(img->data_list[img_depth]);
	  img->data_list[img_depth]=tempdata;
	  img->data_length[img_depth]=newWidth*newHeight;
	}
    }
  img->x_size=newWidth;
  img->y_size=newHeight;
  free(spline_img);
#endif
#ifndef USE_SPLINElib
  fprintf(stderr,"Not compiled with spline support.\n");
#endif
  return(0);
}


int main (int argc, char **argv)
{
  struct arguments arguments;
  IMG_list* img_list;
  int img_nr;
  double* temp_img;
  int width, height;
  
  /* Set argument defaults */
  arguments.mode = scale_BRESENHAM;
  arguments.width = 0;
  arguments.height = 0;
  arguments.args[0] = "";
  arguments.args[1] = "";
  arguments.splinedegree=3;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  img_list=IMG_load_list (arguments.args[0]);


  switch(arguments.mode)
    {
    case scale_BRESENHAM:
      fprintf(stdout,"#The scale algorithm is Bresenham\n");
      break;
    case scale_SPLINE:
      fprintf(stdout,"#The scale algorithm is spline with a splinedegree of %d\n",arguments.splinedegree);
      break;
    }

  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    {
      width=arguments.width;
      height=arguments.height;
      if(width==0)
	{
	  width=(int)(((double)(height*img_list->img[img_nr].x_size)
		       /(double)img_list->img[img_nr].y_size)+0.5);
	  if(width<1)
	    {
	      fprintf(stderr,"calculated width less 1\n");
	      exit(1);
	    }
	}
      if(height==0)
	{
	  height=(int)(((double)(width*img_list->img[img_nr].y_size)
			/(double)img_list->img[img_nr].x_size)+0.5);
	  if(height<1)
	    {
	      fprintf(stderr,"calculated height less 1\n");
	      exit(1);
	    }	  
	}

      switch(arguments.mode)
	{
	case scale_BRESENHAM:
	  img_scale_bresenham( &(img_list->img[img_nr]), height, width);
	  break;
	case scale_SPLINE:
	  img_scale_spline( &(img_list->img[img_nr]), height, width, arguments.splinedegree);
	  break;
	}
    }
  

  IMG_save_list(img_list,arguments.args[1]);
  IMG_free_list(img_list);

  return 0;
}
